from .sql import SqlTask
from .task import TaskStatus


class AutoSqlTask(SqlTask):
    def setup(self):
        self.db = self.sayn_config.default_db

        status = self._setup_file_name()
        if status != TaskStatus.READY:
            return self.failed()

        self.template = self._get_query_template()
        if self.template is None:
            return self.failed()

        status = self._setup_materialisation()
        if status != TaskStatus.READY:
            return status

        status = self._setup_destination()
        if status != TaskStatus.READY:
            return status

        status = self._setup_ddl()
        if status != TaskStatus.READY:
            return status

        status = self._setup_sql()
        if status != TaskStatus.READY:
            return status

        return self._check_extra_fields()

    # Task property methods

    def _setup_materialisation(self):
        # Type of materialisation
        self.materialisation = self._pop_property("materialisation")
        if self.materialisation is None:
            return self.failed(
                '"materialisation" is a required field (values: table, incremental, view)'
            )
        elif not isinstance(self.materialisation, str) or self.materialisation not in (
            "table",
            "incremental",
            "view",
        ):
            return self.failed(
                'Accepted "materialisation" values: table, incremental, view)'
            )
        elif self.materialisation == "incremental":
            self.delete_key = self._pop_property("delete_key")
            if self.delete_key is None:
                return self.failed("Incremental materialisation requires delete_key")

        return TaskStatus.READY

    def _setup_sql(self):
        try:
            query = self.template.render(**self.parameters)
        except Exception as e:
            return self.failed(f"Error compiling template\n{e}")

        if self.ddl is not None and "permissions" in self.ddl:
            permissions = "\n\n-- Grant permissions\n" + self.db.grant_permissions(
                self.table, self.schema, self.ddl["permissions"]
            )
        else:
            permissions = ""

        if self.materialisation == "view":
            # Views just replace the current object if it exists
            self.compiled = (
                self.db.create_table_select(
                    self.table, self.schema, query, replace=True, view=True,
                )
                + permissions
            )
            return self.ready()

        # Some common statements
        tmp_table = self.db.create_table_select(
            self.tmp_table, self.tmp_schema, query, replace=True,
        )

        move = (
            self.db.move_table(
                self.tmp_table,
                self.tmp_schema,
                self.table,
                self.schema,
                self.ddl,
            )
            + permissions
        )

        if not self.db.table_exists(self.table, self.schema):
            # When the table doesn't currently exists, regardless of materialisation, just create it
            if self.ddl is not None and "columns" in self.ddl:
                # create table with DDL and insert the output of the select
                create_table_ddl = (
                    self.db.create_table_ddl(
                        self.table,
                        self.schema,
                        self.ddl,
                        replace=self.sayn_config.options["full_load"],
                    )
                    + permissions
                )
                insert = self.db.insert(self.table, self.schema, query)
                self.compiled = (
                    f"-- Create table\n"
                    f"{create_table_ddl}\n\n"
                    f"-- Load data\n"
                    f"{insert}\n\n"
                )

            else:
                # or create from the select if DDL not present
                self.compiled = (
                    self.db.create_table_select(
                        self.table,
                        self.schema,
                        query,
                        replace=self.sayn_config.options["full_load"],
                    )
                    + permissions
                )

        # If the table exists and there's DDL defined, ...
        elif (
            self.materialisation == "table"
            and self.ddl is not None
            and "columns" in self.ddl
        ):
            # ... when it's a TABLE materialisation: create, load and replace table with temporary
            create_tmp_ddl = self.db.create_table_ddl(
                self.tmp_table, self.tmp_schema, self.ddl, replace=True,
            )
            insert_tmp = self.db.insert(
                self.tmp_table, self.tmp_schema, query
            )
            self.compiled = (
                f"-- Create temporary table\n"
                f"{create_tmp_ddl}\n\n"
                f"-- Load data\n"
                f"{insert_tmp}\n\n"
                f"-- Move data to final table\n"
                f"{move}"
            )

        elif self.materialisation == "table" and (
            self.ddl is None or "columns" not in self.ddl
        ):
            # ... when it exists and it's a TABLE materialisation, create in temp and move
            self.compiled = (
                f"-- Create temporary table\n"
                f"{tmp_table}\n\n"
                f"-- Move data to final table\n"
                f"{move}"
            )

        elif self.materialisation == "incremental":
            # ... whether there's DDL or not, ...
            # ... when it's a INCREMENTAL materialisation: MERGE instead of load
            merge = self.db.merge_tables(
                self.tmp_table,
                self.tmp_schema,
                self.table,
                self.schema,
                self.delete_key,
            )
            self.compiled = (
                f"-- Create temporary table\n"
                f"{tmp_table}\n\n"
                f"-- Move data to final table\n"
                f"{merge}\n\n"
            )

        else:
            raise ValueError("SAYN issue")

        return TaskStatus.READY



import logging

from .task import Task


class DummyTask(Task):
    def setup(self):
        logging.debug("Setting up Dummy Task")
        return self.ready()

    def compile(self):
        logging.debug("DummyTask compiling")
        return self.success()

    def run(self):
        logging.debug("DummyTask running.")
        return self.success()



from .postgresql import Postgresql


class Redshift(Postgresql): pass
#     def __init__(self, name, name_in_settings, settings):
#         # self.dialect = "postgresql"
#         # connection_details = settings
#         # connection_details.pop("type")
#         super().__init__(name, name_in_settings, connection_details)



from datetime import datetime, date, timedelta

import logging
import sys

import click

from ..config import Config, SaynConfigError
from ..utils.logger import Logger
from ..dag import Dag
from ..start_project.start_project import sayn_init

click_debug = click.option(
    "--debug", "-d", is_flag=True, default=False, help="Include debug messages"
)


def click_filter(func):
    func = click.option(
        "--tasks",
        "-t",
        multiple=True,
        help="Task query to INCLUDE in the execution: [+]task_name[+], dag:dag_name, tag:tag_name",
    )(func)
    func = click.option(
        "--exclude",
        "-x",
        multiple=True,
        help="Task query to EXCLUDE in the execution: [+]task_name[+], dag:dag_name, tag:tag_name",
    )(func)
    return func


def click_incremental(func):
    func = click.option(
        "--full-load", "-f", is_flag=True, default=False, help="Do a full load"
    )(func)
    func = click.option(
        "--start-dt",
        "-s",
        # type=click.DateTime(formats=["%Y-%m-%d"]),
        default=str(date.today() + timedelta(days=-1)),
        help="For incremental loads, the start date",
    )(func)
    func = click.option(
        "--end-dt",
        "-e",
        # type=click.DateTime(formats=["%Y-%m-%d"]),
        default=str(date.today() + timedelta(days=-1)),
        help="For incremental loads, the end date",
    )(func)
    return func


def click_run_options(func):
    func = click_debug(func)
    func = click.option("--profile", "-p", help="Profile from settings to use")(func)
    func = click_incremental(func)
    func = click_filter(func)
    return func


@click.group(help="SAYN management tool.")
def cli():
    pass


# Click commands


@cli.command(help="Initialise a SAYN project in working directory.")
@click.argument("sayn_project_name")
def init(sayn_project_name):
    sayn_init(sayn_project_name)


@cli.command(help="Compile sql tasks.")
@click_run_options
def compile(debug, tasks, exclude, profile, full_load, start_dt, end_dt):
    run_command("compile", debug, tasks, exclude, profile, full_load, start_dt, end_dt)


@cli.command(help="Run SAYN tasks.")
@click_run_options
def run(debug, tasks, exclude, profile, full_load, start_dt, end_dt):
    run_command("run", debug, tasks, exclude, profile, full_load, start_dt, end_dt)


def run_command(command, debug, tasks, exclude, profile, full_load, start_dt, end_dt):
    run_start_ts = datetime.now()
    run_id = (run_start_ts - datetime(1970, 1, 1)).total_seconds()

    Logger(run_id=run_id, debug=debug)

    try:
        Config(
            profile=profile, full_load=full_load, start_dt=start_dt, end_dt=end_dt,
        )
    except SaynConfigError as e:
        logging.error(e)
        sys.exit(1)

    dag = Dag(tasks_query=tasks, exclude_query=exclude)

    if command == "run":
        dag.run()
    elif command == "compile":
        dag.compile()
    else:
        raise ValueError(f'Unknown command "{command}"')

    logging.info(f"{command.capitalize()} took {datetime.now() - run_start_ts}")


@cli.command(help="Generate DAG image")
@click_debug
@click_filter
def dag_image(debug, tasks, exclude):
    run_start_ts = datetime.now()
    run_id = (run_start_ts - datetime(1970, 1, 1)).total_seconds()

    Logger(run_id=run_id, debug=debug)

    try:
        Config()
    except SaynConfigError as e:
        logging.error(e)
        sys.exit(1)

    dag = Dag(tasks_query=tasks, exclude_query=exclude)
    dag.plot(folder="images", file_name="dag")



from itertools import groupby

from .database import Database


class Postgresql(Database):
    def __init__(self, name, name_in_settings, settings):
        self.dialect = "postgresql"
        connection_details = settings
        connection_details.pop("type")
        super().__init__(name, name_in_settings, connection_details)

    def create_table_select(self, table, schema, select, replace=False, view=False):
        table = f"{schema+'.' if schema else ''}{table}"
        table_or_view = "VIEW" if view else "TABLE"

        q = ""
        if replace:
            q += f"DROP {table_or_view} IF EXISTS {table} CASCADE;\n"
        q += f"CREATE {table_or_view} {table} AS (\n{select}\n);"

        return q

    def create_table_ddl(self, table, schema, ddl, replace=False):
        table_name = table
        table = f"{schema+'.' if schema else ''}{table_name}"
        columns = "\n    , ".join(
            [
                (
                    f'{c["name"]} {c["type"]}'
                    f'{" PRIMARY KEY" if c.get("primary", False) else ""}'
                    f'{" NOT NULL" if c.get("not_null", False) else ""}'
                )
                for c in ddl["columns"]
            ]
        )
        indexes = {
            idx: [c[1] for c in cols]
            for idx, cols in groupby(
                sorted(
                    [
                        (idx, c["name"])
                        for c in ddl["columns"]
                        if "indexes" in c
                        for idx in c["indexes"]
                    ]
                ),
                lambda x: x[0],
            )
        }

        q = ""
        if replace:
            q += f"DROP TABLE IF EXISTS {table} CASCADE;\n"
        q += f"CREATE TABLE {table} (\n      {columns}\n);\n"
        q += "\n".join(
            [
                f"CREATE INDEX {table_name}_{name} ON {table}({', '.join(cols)});"
                for name, cols in indexes.items()
            ]
        )

        return q

    def drop_table(self, table, schema, view=False):
        table = f"{schema+'.' if schema else ''}{table}"
        table_or_view = "VIEW" if view else "TABLE"
        return f"DROP {table_or_view} IF EXISTS {table} CASCADE;"

    def move_table(self, src_table, src_schema, dst_table, dst_schema, ddl):
        drop = f"DROP TABLE IF EXISTS {dst_schema+'.' if dst_schema else ''}{dst_table} CASCADE;"
        rename = f"ALTER TABLE {src_schema+'.' if src_schema else ''}{src_table} RENAME TO {dst_table};"
        if dst_schema is not None and dst_schema != src_schema:
            change_schema = f"ALTER TABLE {src_schema+'.' if src_schema else ''}{dst_table} SET SCHEMA {dst_schema};"
        else:
            change_schema = ""

        idx_alter = []
        if ddl is not None and "columns" in ddl:
            # Change primary key name
            if len([c for c in ddl["columns"] if c.get("primary", False)]):
                idx_alter.append(
                    f"ALTER INDEX {dst_schema+'.' if dst_schema else ''}{src_table}_pkey RENAME TO {dst_table}_pkey;"
                )

            # Change index names
            for idx in [
                idx for c in ddl["columns"] for idx in c.get("indexes", list())
            ]:
                idx_alter.append(
                    f"ALTER INDEX {dst_schema+'.' if dst_schema else ''}{src_table}_{idx} RENAME TO {dst_table}_{idx};"
                )

        return "\n".join([drop, rename, change_schema] + idx_alter)



from copy import copy, deepcopy
from collections import deque, OrderedDict
from datetime import datetime
from itertools import groupby
import logging
import sys

from .config import Config
from .utils.singleton import singleton
from .utils.logger import Logger
from .tasks import create_task, TaskStatus, IgnoreTask


class DagValidationError(Exception):
    pass


@singleton
class Dag:
    """Directed acyclic graph implementation."""

    graph = OrderedDict()
    tasks = dict()

    def __init__(self, tasks_query=(), exclude_query=()):
        Logger().set_config(stage="DAG")
        logging.info("----------")
        logging.info(f"Setting up. Run ID: {Config().run_id}")
        logging.info("----------")

        task_definitions = Config()._task_definitions

        # Setup DAG structure using just names and parents
        self._from_dict_reversed(
            {k: v.get("parents", list()) for k, v in task_definitions.items()}
        )

        # Calculate list of tasks in the correct order of execution based on parentage
        self.tasks = {name: None for name in self.topological_sort()}

        # Get the list of tasks that need to run in this execution

        # 1. Get a dict of tags > list of tasks with that tag (used by _task_query to get the list of relevant tasks)
        tags = {
            tag: [i[1] for i in tag_tasks]
            for tag, tag_tasks in groupby(
                sorted(
                    [
                        (tag, task_name)
                        for task_name, task_def in task_definitions.items()
                        for tag in task_def.get("tags", list())
                    ],
                    key=lambda x: x[0],
                ),
                lambda x: x[0],
            )
        }

        # 2. Get a dict of dags > task list with that model (used by _task_query to get the list of relevant tasks)
        dags = {
            dag: [i[1] for i in dag_tasks]
            for dag, dag_tasks in groupby(
                sorted(
                    [
                        (task_def["dag"], task_name)
                        for task_name, task_def in task_definitions.items()
                    ],
                    key=lambda x: x[0],
                ),
                lambda x: x[0],
            )
        }

        # 3. We generate tasks_to_process containing all tasks to be run
        if len(tasks_query) == 0:
            # If there no task filter, we'll run all tasks
            tasks_to_process = set(self.tasks.keys())
        else:
            # Otherwise we get the list of tasks corresponding to what's specified in --tasks
            tasks_to_process = set(
                t for q in tasks_query for t in self._task_query(tags, dags, q)
            )

        # We always apply the exclude filter query
        tasks_to_process = tasks_to_process - set(
            t for q in exclude_query for t in self._task_query(tags, dags, q)
        )

        # Create task objects and set them up
        self.tasks = {
            name: create_task(
                name,
                task_definitions[name]["type"],
                task_definitions[name],
                name not in tasks_to_process,
            )
            for name in self.tasks.keys()
        }

        # Once all objects are created, we can add references to those in each task
        for _, task in self.tasks.items():
            task.set_parents(self.tasks)

        # Run the setup for each task
        for _, task in self.tasks.items():
            Logger().set_config(task=task.name)
            if task.status == TaskStatus.FAILED:
                continue
            task.setup()

        # delete the task_definitions attribute from the config so it we do not have tasks on both APIs
        # delattr(Config(), "_task_definitions")
        Logger().set_config(task=None)
        logging.info("DAG Setup: done.")

    def _task_query(self, tags, dags, query):
        """Returns a list of tasks from the dag matching the query"""

        if query[0] == "+":
            # A "+" before a task name means the specified task and all its parents
            task = query[1:]
            if task not in self.tasks:
                raise KeyError(f'Tag "{task}" not in dag')
            return [task] + self.all_upstreams(task)

        elif query[-1] == "+":
            # A "+" after a task name means the specified task and all its children
            task = query[:-1]
            if task not in self.tasks:
                raise KeyError(f'Task "{task}" not in dag')
            return [task] + self.all_downstreams(task)

        elif query[:4] == "tag:":
            # A tag name will be specified as `tag:tag_name`
            tag = query[4:]
            if tag not in tags:
                raise KeyError(f'Tag "{tag}" not in dag')
            return tags[tag]

        elif query[:4] == "dag:":
            # A dag name will be specified as `dag:dag_name`
            dag = query[4:]
            if dag not in dags:
                raise KeyError(f'DAG "{dag}" does not exists')
            return dags[dag]

        elif query in self.tasks:
            # Otherwise it should be a single task name
            return [query]

        else:
            # ... or the task doesn't exists in the dag
            raise KeyError(f'Task "{query}" not in dag')

    def _run_task(self, command, task):
        Logger().set_config(task=task.name)
        task_start_ts = datetime.now()
        if task.status != TaskStatus.READY:
            task.failed("Task failed during setup. Skipping...")
        elif not task.can_run():
            logging.warn("SKIPPING")
            task.skipped()
        else:
            logging.debug("Starting")
            task.executing()
            try:
                if command == "compile":
                    status = task.compile()
                elif command == "run":
                    status = task.run()
                else:
                    status = None
            except Exception as e:
                logging.exception(e)

                status = None

            if status is None:
                task.status = TaskStatus.UNKNOWN
                logging.error(
                    f"Finished in an unknown state ({datetime.now() - task_start_ts})"
                )
            elif status != TaskStatus.SUCCESS:
                logging.error(f"Failed status ({datetime.now() - task_start_ts})")
            else:
                logging.info(
                    f"\u001b[32mSuccess ({datetime.now() - task_start_ts})\u001b[0m"
                )

        return task.status

    def _run_command(self, command):
        Logger().set_config(stage="Run", task=None, progress="0")
        logging.info("----------")
        logging.info(f"Running. Run ID: {Config().run_id}")
        logging.info("----------")

        failed = list()
        success = list()
        skipped = list()

        tasks_to_run = [
            task for task in self.tasks.values() if not isinstance(task, IgnoreTask)
        ]
        ntasks = len(tasks_to_run)

        tcounter = 0
        for task in self.tasks.values():
            if not task.should_run():
                # For IgnoreTasks
                task.success()
                continue

            status = self._run_task(command, task)
            if status == TaskStatus.SKIPPED:
                skipped.append(task.name)
            elif status in (TaskStatus.FAILED, TaskStatus.UNKNOWN) or status != TaskStatus.SUCCESS:
                failed.append(task.name)
            else:
                success.append(task.name)

            tcounter += 1
            run_progress = str(round((tcounter) / ntasks * 100))
            Logger().set_config(progress=run_progress)

        Logger().set_config(stage="Summary", task=None, progress=None)
        logging.info("----------")
        logging.info(f"Stats for Run ID: {Config().run_id}")
        logging.info("----------")

        recap_str = list()
        recap_str.append(
            f"Process finished. Total tasks: {ntasks}. Success: {len(success)}. Failed {len(failed)}. Skipped {len(skipped)}."
        )
        if len(success) > 0:
            recap_str.append(f"The following tasks succeded: {', '.join(success)}")
        if len(failed) > 0:
            recap_str.append(f"The following tasks failed: {', '.join(failed)}")
        if len(skipped) > 0:
            recap_str.append(f"The following tasks were skipped: {', '.join(skipped)}")

        if len(failed) > 0:
            log_func = logging.error
        elif len(skipped) > 0:
            log_func = logging.warning
        else:
            log_func = lambda x: logging.info(f"\u001b[32m{x}\u001b[0m")

        for msg in recap_str:
            log_func(msg)

    def run(self):
        self._run_command("run")

    def compile(self):
        self._run_command("compile")

    def plot(self, folder=None, file_name=None):
        """Uses graphviz to plot the dag
        It requires the graphviz python package (pip install graphviz) and an installation of graphviz
        (eg: brew install graphviz)
        """

        task_list = list(self.tasks.keys())

        try:
            from graphviz import Digraph
        except:
            logging.error(
                "Graphviz is required. To install it `pip install graphviz` and install it in your system (eg: `brew install graphviz"
            )
            sys.exit()
        dot = Digraph(comment="SAYN", graph_attr={"splines": "ortho", "nodesep": "0.8"})
        for n in task_list:
            dot.node(n, style="rounded", shape="box")
        for a, l in self.graph.items():
            if a in task_list:
                for b in l:
                    if b in task_list:
                        dot.edge(a, b)

        if file_name is not None:
            dot.render(
                directory=folder,
                filename=file_name,
                cleanup=True,
                view=False,
                format="png",
            )

    # Creation

    def _from_dict(self, graph_dict):
        """Reset the graph and build it from the passed dictionary.
        The dictionary takes the form of {node_name: [directed edges]}
        """

        self.reset_graph()

        graph_dict = deepcopy(graph_dict)
        graph_dict.update(
            {v: [] for l in graph_dict.values() for v in l if v not in graph_dict}
        )

        for new_node in graph_dict.keys():
            self.add_node(new_node)
        for ind_node, dep_nodes in graph_dict.items():
            if not isinstance(dep_nodes, list):
                raise TypeError("dict values must be lists")
            for dep_node in dep_nodes:
                self.add_edge(ind_node, dep_node)

    def _from_dict_reversed(self, graph_dict):
        """Reverses the dictionary before calling _from_dict"""
        self._from_dict(self._reverse_dict(graph_dict))

    def _reverse_dict(self, graph_dict):
        """Reverses the dictionary"""
        out = {
            k: list([v[1] for v in g])
            for k, g in groupby(
                sorted([(v, a) for a, l in graph_dict.items() for v in l]),
                key=lambda x: x[0],
            )
        }
        # Make sure all nodes are in
        out.update(
            {
                k: []
                for k in set(
                    [v for a, l in graph_dict.items() for v in l if v not in out]
                    + [k for k in graph_dict.keys() if k not in out]
                )
            }
        )
        return out

    def reset_graph(self):
        """Restore the graph to an empty state."""
        self.graph = OrderedDict()

    def validate(self, graph=None):
        """Returns (Boolean, message) of whether DAG is valid."""
        graph = graph if graph is not None else self.graph
        if len(self.ind_nodes(graph)) == 0:
            return (False, "no independent nodes detected")
        try:
            self.topological_sort(graph)
        except ValueError:
            return (False, "failed topological sort")
        return (True, "valid")

    # Manipulation

    def add_node(self, node_name, graph=None):
        """Add a node if it does not exist yet, or error out."""
        if not graph:
            graph = self.graph
        if node_name in graph:
            raise KeyError("node %s already exists" % node_name)
        graph[node_name] = set()

    def add_node_if_not_exists(self, node_name, graph=None):
        try:
            self.add_node(node_name, graph=graph)
        except KeyError:
            pass

    def delete_node(self, node_name, graph=None):
        """Deletes this node and all edges referencing it."""
        if not graph:
            graph = self.graph
        if node_name not in graph:
            raise KeyError("node %s does not exist" % node_name)
        graph.pop(node_name)

        for node, edges in iter(graph.items()):
            if node_name in edges:
                edges.remove(node_name)

    def delete_node_if_exists(self, node_name, graph=None):
        try:
            self.delete_node(node_name, graph=graph)
        except KeyError:
            pass

    def add_edge(self, ind_node, dep_node, graph=None):
        """Add an edge (dependency) between the specified nodes."""
        if not graph:
            graph = self.graph
        if ind_node not in graph or dep_node not in graph:
            raise KeyError("one or more nodes do not exist in graph")
        test_graph = deepcopy(graph)
        test_graph[ind_node].add(dep_node)
        is_valid, message = self.validate(test_graph)
        if is_valid:
            graph[ind_node].add(dep_node)
        else:
            raise DagValidationError()

    def delete_edge(self, ind_node, dep_node, graph=None):
        """Delete an edge from the graph."""
        if not graph:
            graph = self.graph
        if dep_node not in graph.get(ind_node, []):
            raise KeyError("this edge does not exist in graph")
        graph[ind_node].remove(dep_node)

    def rename_edges(self, old_task_name, new_task_name, graph=None):
        """Change references to a task in existing edges."""
        if not graph:
            graph = self.graph
        for node, edges in graph.items():

            if node == old_task_name:
                graph[new_task_name] = copy(edges)
                del graph[old_task_name]

            else:
                if old_task_name in edges:
                    edges.remove(old_task_name)
                    edges.add(new_task_name)

    # Navigation

    def predecessors(self, node, graph=None):
        """Returns a list of all predecessors of the given node."""
        if graph is None:
            graph = self.graph
        return [key for key in graph if node in graph[key]]

    def upstream(self, node, graph=None):
        """Returns a list of all nodes this node has edges towards."""
        if graph is None:
            graph = self.graph
        if node not in graph:
            raise KeyError("node %s is not in graph" % node)
        return list(self.predecessors(node, graph))

    def all_upstreams(self, node, graph=None):
        """Returns a list of all nodes ultimately downstream
        of the given node in the dependency graph, in
        topological order.
        """
        if graph is None:
            graph = self.graph
        nodes = [node]
        nodes_seen = set()
        i = 0
        while i < len(nodes):
            upstreams = self.upstream(nodes[i], graph)
            for upstream_node in upstreams:
                if upstream_node not in nodes_seen:
                    nodes_seen.add(upstream_node)
                    nodes.append(upstream_node)
            i += 1
        return list(
            filter(lambda node: node in nodes_seen, self.topological_sort(graph=graph))
        )

    def downstream(self, node, graph=None):
        """Returns a list of all nodes this node has edges towards."""
        if graph is None:
            graph = self.graph
        if node not in graph:
            raise KeyError("node %s is not in graph" % node)
        return list(graph[node])

    def all_downstreams(self, node, graph=None):
        """Returns a list of all nodes ultimately downstream
        of the given node in the dependency graph, in
        topological order.
        """
        if graph is None:
            graph = self.graph
        nodes = [node]
        nodes_seen = set()
        i = 0
        while i < len(nodes):
            downstreams = self.downstream(nodes[i], graph)
            for downstream_node in downstreams:
                if downstream_node not in nodes_seen:
                    nodes_seen.add(downstream_node)
                    nodes.append(downstream_node)
            i += 1
        return list(
            filter(lambda node: node in nodes_seen, self.topological_sort(graph=graph))
        )

    def all_leaves(self, graph=None):
        """Return a list of all leaves (nodes with no downstreams)"""
        if graph is None:
            graph = self.graph
        return [key for key in graph if not graph[key]]

    def ind_nodes(self, graph=None):
        """Returns a list of all nodes in the graph with no dependencies."""
        if graph is None:
            graph = self.graph

        dependent_nodes = set(
            node for dependents in iter(graph.values()) for node in dependents
        )
        return [node for node in graph.keys() if node not in dependent_nodes]

    def topological_sort(self, graph=None):
        """Returns a topological ordering of the DAG.
        Raises an error if this is not possible (graph is not valid).
        """
        if graph is None:
            graph = self.graph

        in_degree = {}
        for u in graph:
            in_degree[u] = 0

        for u in graph:
            for v in graph[u]:
                in_degree[v] += 1

        queue = deque()
        for u in in_degree:
            if in_degree[u] == 0:
                queue.appendleft(u)

        l = []
        while queue:
            u = queue.pop()
            l.append(u)
            for v in graph[u]:
                in_degree[v] -= 1
                if in_degree[v] == 0:
                    queue.appendleft(v)

        if len(l) == len(graph):
            return l
        else:
            raise ValueError("graph is not acyclic")

    def size(self):
        return len(self.graph)



import logging

from .task import Task


class IgnoreTask(Task):
    def should_run(self):
        return False

    def setup(self):
        logging.debug("Setting up IgnoreTask")
        return self.ready()

    def compile(self):
        logging.debug("IgnoreTask compiling")
        return self.success()

    def run(self):
        logging.debug("IgnoreTask running.")
        return self.success()



import copy
from datetime import datetime, date, timedelta
import importlib
import json
import logging
import os
from pathlib import Path
import shutil
import sys

from jinja2 import Environment, FileSystemLoader, StrictUndefined

from .utils import yaml
from .utils.singleton import singleton
from . import database


class SaynConfigError(Exception):
    pass


@singleton
class Config:
    _task_definitions = None

    def __init__(
        self,
        profile=None,
        full_load=False,
        start_dt=None,
        end_dt=None,
        **cmd_parameters,
    ):
        # Process basic command parameters
        self.run_id = (datetime.now() - datetime(1970, 1, 1)).total_seconds()

        # For dates we default to yesterday
        yesterday = date.today() - timedelta(days=1)
        if start_dt is None:
            start_dt = yesterday
        elif type(start_dt) == str:
            start_dt = datetime.strptime(start_dt, "%Y-%m-%d")

        if end_dt is None:
            end_dt = yesterday
        elif type(end_dt) == str:
            end_dt = datetime.strptime(end_dt, "%Y-%m-%d")

        if end_dt < start_dt:
            raise SaynConfigError("end-dt can't be earlier than start-dt")

        self.options = {
            "full_load": full_load,
            "start_dt": start_dt,
            "end_dt": end_dt,
        }

        # Read the config
        config = self._read_config(profile)
        if config is None:
            raise SaynConfigError("Error detected in configuration")

        # Process and store the config from the yaml files
        self.options.update(
            {"selected_profile": config["selected_profile"],}
        )

        self._setup_parameters(config["parameters"], cmd_parameters)

        self._setup_credentials(config["default_db"], config["credentials"])

        self._setup_tasks(config["presets"], config["dags"])
        if self._task_definitions is None:
            raise SaynConfigError(f"Error reading tasks")

    ###############################
    # Reader methods
    ###############################

    def _read_config(self, profile):
        project = self._read_project()
        if project is None:
            return

        settings = self._read_settings(
            profile,
            project["required_credentials"],
            list(project["parameters"].keys()),
        )
        if settings is None:
            return

        parameters = project["parameters"]
        parameters.update(settings["parameters"])

        return {
            "selected_profile": settings["selected_profile"],
            "parameters": parameters,
            "default_db": project["default_db"],
            "credentials": settings["credentials"],
            "presets": project["presets"],
            "dags": project["dags"],
        }

    def _read_project(self):
        path = Path("project.yaml")
        parsed = yaml.load(path)
        if parsed is None:
            return

        dags = {g.name[: -len(g.suffix)]: None for g in Path("dags").glob("*.yaml")}

        schema = yaml.Map(
            {
                yaml.Optional("parameters"): yaml.MapPattern(
                    yaml.NotEmptyStr(), yaml.Any()
                ),
                "required_credentials": yaml.UniqueSeq(yaml.NotEmptyStr()),
                yaml.Optional("default_db"): yaml.NotEmptyStr(),
                yaml.Optional("presets"): yaml.MapPattern(
                    yaml.Identifier(), yaml.MapPattern(yaml.NotEmptyStr(), yaml.Any()),
                ),
                "dags": yaml.UniqueSeq(yaml.Enum(dags.keys())),
            }
        )

        try:
            parsed.revalidate(schema)
        except yaml.ValidationError as e:
            logging.error(f"Error reading {path.name}")
            logging.error(e)
            return

        self._setup_paths()

        # Read dags
        dags = {name: self._read_dag(name) for name in dags.keys()}
        if len([v for v in dags.values() if v is None]) > 0:
            # If any errors reading dags, fail the reading
            return

        # Determine default_db
        default_db = parsed.data.get("default_db")

        if default_db is None and len(parsed["required_credentials"]) == 1:
            default_db = parsed["required_credentials"].data[0]
        elif default_db is None and len(parsed["required_credentials"]) > 1:
            logging.error(
                "Default database is required when multiple credentials are set"
            )
            return
        elif default_db not in parsed["required_credentials"]:
            logging.error(f'Default database "{default_db}" not defined')
            return
        # else:
        #    default_db already points to an existing credential

        return {
            "parameters": parsed.data.get("parameters", dict()),
            "required_credentials": parsed.data["required_credentials"],
            "default_db": default_db,
            "presets": parsed.data.get("presets", dict()),
            "dags": dags,
        }

    def _read_dag(self, dag_name):
        path = Path(self.dags_path, f"{dag_name}.yaml")
        parsed = yaml.load(path)
        if parsed is None:
            return

        schema = yaml.Map(
            {
                yaml.Optional("presets"): yaml.MapPattern(
                    yaml.Identifier(), yaml.MapPattern(yaml.NotEmptyStr(), yaml.Any()),
                ),
                "tasks": yaml.MapPattern(
                    yaml.Identifier(), yaml.MapPattern(yaml.NotEmptyStr(), yaml.Any()),
                ),
            }
        )

        try:
            parsed.revalidate(schema)
        except yaml.ValidationError as e:
            logging.error(f'Error reading dag "{path.name}"')
            logging.error(e)
            return

        return {
            "presets": parsed.data.get("presets", dict()),
            "tasks": parsed.data.get("tasks", dict()),
        }

    def _read_settings(self, profile, required_credentials, allowed_parameters):
        env_vars = {k: v for k, v in os.environ.items() if k.startswith("SAYN_")}
        if len(env_vars) > 0:
            print("Reading settings from environment variables")
            return self._read_settings_environment(
                env_vars, required_credentials, allowed_parameters
            )
        else:
            print('Reading settings from "settings.yaml"')
            return self._read_settings_file(
                profile, required_credentials, allowed_parameters
            )

    def _read_settings_environment(
        self, env_vars, required_credentials, allowed_parameters
    ):
        parameters = dict()
        credentials = dict()

        for key, value in env_vars.items():
            if key.startswith("SAYN_PARAMETER_"):
                parameter_key = key[len("SAYN_PARAMETER_") :]
                if parameter_key not in allowed_parameters:
                    logging.error(
                        f'Parameter "{parameter_key}" not defined in models.yaml. '
                    )
                    logging.error(
                        f"Allowed parameters are {', '.join(allowed_parameters)}"
                    )
                    return
                try:
                    # Try to interpret it as a json
                    parameters[key] = json.loads(value)
                except:
                    # Treat it as a string
                    parameters[parameter_key] = value

            elif key.startswith("SAYN_CREDENTIAL_"):
                credential_key = key[len("SAYN_CREDENTIAL_") :]
                if credential_key not in required_credentials:
                    continue
                try:
                    # Needs to be a json as the base structure for credentials is a map
                    credentials[credential_key] = {
                        "name_in_settings": None,
                        "settings": json.loads(value),
                    }
                except Exception as e:
                    logging.error(
                        f'Error trying to parse environment variable "{key}" as json'
                    )
                    logging.error(e)
                    return

            else:
                logging.error(f'Environment variable "{key}" is not recognised')
                return

        # Ensure all credentials are specified
        for credential in required_credentials:
            if credential not in credentials:
                logging.error(f'Missing credential "{credential}"')
                return

        return {
            "selected_profile": None,
            "parameters": parameters,
            "credentials": credentials,
        }

    def _read_settings_file(self, profile, required_credentials, allowed_parameters):
        path = Path("settings.yaml")
        parsed = yaml.load(path)
        if parsed is None:
            return

        schema = yaml.Map(
            {
                yaml.Optional("default_profile"): yaml.NotEmptyStr(),
                "profiles": yaml.MapPattern(
                    yaml.NotEmptyStr(),  # profile name
                    yaml.Map(
                        {
                            "credentials": yaml.Map(
                                {c: yaml.NotEmptyStr() for c in required_credentials}
                            ),
                            yaml.Optional("parameters"): yaml.MapPattern(
                                yaml.Enum(allowed_parameters), yaml.Any()
                            ),
                        }
                    ),
                ),
                "credentials": yaml.MapPattern(
                    yaml.NotEmptyStr(), yaml.MapPattern(yaml.NotEmptyStr(), yaml.Any())
                ),
            }
        )

        try:
            parsed.revalidate(schema)
        except yaml.ValidationError as e:
            logging.error(f'Error reading settings "{path.name}"')
            logging.error(e)
            return

        # Extract values on first pass
        selected_profile = profile or parsed.data.get("default_profile")

        # Get the default_profile
        if selected_profile is None and len(parsed["profiles"]) > 1:
            logging.error("Default profile is required when multiple profiles are set")
            return
        elif selected_profile is None and len(parsed["profiles"]) == 1:
            selected_profile = list(parsed["profiles"].data.keys())[0]
        elif selected_profile not in parsed["profiles"]:
            logging.error(f'Default profile "{selected_profile}" not defined')
            return

        profile = parsed["profiles"][selected_profile]

        # Keeping only the credentials specified in the profile
        credentials = {
            n: {"name_in_settings": n, "settings": c}
            for n, c in parsed.data["credentials"].items()
        }
        no_type = [n for n, c in credentials.items() if "type" not in c["settings"]]
        if len(no_type) > 0:
            logging.error(f"Some credentials have no type: {', '.join(no_type)}")
            return
        credentials = {
            name: credentials[name_in_settings]
            for name, name_in_settings in profile.data["credentials"].items()
        }

        return {
            "selected_profile": selected_profile,
            "parameters": profile.data.get("parameters", dict()),
            "credentials": credentials,
        }

    ###############################
    # Setup config object methods
    ###############################

    def _setup_paths(self, **paths):
        self.dags_path = Path(paths.get("dags", "dags"))
        self.sql_path = Path(paths.get("sql", "sql"))
        self.python_path = Path(paths.get("python", "python"))

        self.compile_path = Path(paths.get("compile", "compile"))
        if self.compile_path.exists():
            if self.compile_path.is_dir():
                shutil.rmtree(self.compile_path.absolute())
            else:
                self.compile_path.unlink()

    def _setup_parameters(self, from_config, from_cmd):
        self.parameters = from_config
        self.parameters.update(from_cmd)

        start_dt = self.options["start_dt"]
        start_dt = start_dt.strftime("%Y-%m-%d") if type(start_dt) != str else start_dt

        end_dt = self.options["end_dt"]
        end_dt = end_dt.strftime("%Y-%m-%d") if type(end_dt) != str else end_dt

        self.parameters.update(
            {
                "start_dt": f"'{start_dt}'",
                "end_dt": f"'{end_dt}'",
                "full_load": self.options["full_load"],
            }
        )

        self.jinja_env = Environment(
            loader=FileSystemLoader(os.getcwd()),
            undefined=StrictUndefined,
            keep_trailing_newline=True,
        )
        self.jinja_env.globals.update(**self.parameters)

    def _setup_credentials(self, default_db, credentials):
        db_credentials = {
            n: c for n, c in credentials.items() if c["settings"]["type"] != "api"
        }
        self.api_credentials = {
            n: {k: v for k, v in c["settings"].items() if k != "type"}
            for n, c in credentials.items()
            if c["settings"]["type"] == "api"
        }

        # Validate the default_db against db credentials
        if default_db not in db_credentials:
            logging.error(
                f"Credential {credentials[default_db]['name_in_settings']} is not a database"
            )
            return

        self.dbs = database.create_all(db_credentials)
        self.default_db = self.dbs[default_db]

    ###############################
    # Task processing methods
    ###############################

    def _setup_python_tasks_module(self):
        path = Path(self.python_path, "__init__.py")
        if not path.is_file():
            raise SaynConfigError(f"Missing file: {path.fullname}")

        loader = importlib.machinery.SourceFileLoader(
            "sayn_python_tasks", str(Path(self.python_path, "__init__.py"))
        )
        spec = importlib.util.spec_from_loader("sayn_python_tasks", loader)
        m = None
        m = importlib.util.module_from_spec(spec)

        sys.modules[spec.name] = m

        try:
            loader.exec_module(m)
        except Exception as e:
            logging.error(f"Error importing python module")
            logging.error(e)
            raise SaynConfigError(
                f"Error importing python folder {self.python_path.absolute}"
            )

    def _get_dag_presets(self, dag_presets, dag_name, global_presets):
        presets = copy.deepcopy(dag_presets)
        for preset_name, preset in presets.items():
            if preset.get("preset") is not None:
                # If the preset references a preset, embed it
                if preset["preset"] in global_presets:
                    referenced_preset = copy.deepcopy(global_presets[preset["preset"]])

                    # Merge common fields to all task types the outer preset
                    if "parameters" in referenced_preset:
                        if "parameters" not in preset:
                            preset["parameters"] = dict()
                        preset["parameters"].update(
                            {
                                k: v
                                for k, v in referenced_preset.pop("parameters").items()
                                if k not in preset["parameters"]
                            }
                        )

                    for property in ("parents", "tags"):
                        if property in referenced_preset:
                            if property not in preset:
                                preset[property] = list()
                            preset[property].extend(referenced_preset.pop(property))

                    # Add the rest of the global preset to the preset
                    preset["preset"] = referenced_preset

                else:
                    # Global preset not defined
                    logging.error(
                        "Referenced preset {} in dag {} not defined in project.yaml".format(
                            preset.get("preset"), dag_name
                        )
                    )
                    return

        out_presets = copy.deepcopy(global_presets)
        out_presets.update(presets)
        return out_presets

    def _get_dag_tasks(self, tasks, dag_name, presets, all_tasks):
        tasks = copy.deepcopy(tasks)
        for name, task in tasks.items():
            # Add the dag name
            task["dag"] = dag_name

            # Check for duplicates
            if name in all_tasks:
                logging.error(
                    "Duplicate tasks in dag {}: {}".format(
                        dag_name,
                        ", ".join(set(tasks.keys().intersection(set(all_tasks)))),
                    )
                )
                return

            # Embed the preset in the task definition
            if "preset" in task:
                if task["preset"] in presets:
                    task["preset"] = copy.deepcopy(presets[task["preset"]])
                else:
                    logging.error(
                        "Preset {} referenced in task {} not found".format(
                            task["preset"], name
                        )
                    )
                    return

            # Consolidate the type of the task
            if "type" not in task:
                if "type" in task.get("preset", dict()):
                    # Try to take it from the preset
                    task["type"] = task["preset"].pop("type")
                    if "preset" in task.get("preset", dict()):
                        # Cleanup the type from the nested preset if present as well
                        task["preset"]["preset"].pop("type")
                elif "preset" in task.get("preset", dict()):
                    # Nested preset
                    task["type"] = task["preset"]["preset"].pop("type")
                else:
                    logging.error(f"Missing type in task {name} in dag {dag_name}")
                    return

            # Consolidate the parents and tags list
            for property in ("parents", "tags"):
                if property not in task:
                    task[property] = list()

                task[property].extend(
                    [
                        v
                        for v in task.get("preset", dict()).pop(property, list())
                        if v not in task[property]
                    ]
                )

            # Consolidate the parameters from the preset
            if "parameters" not in task:
                task["parameters"] = dict()

            if "parameters" in task.get("preset", dict()):
                # We give priority to the values defined in the task, so we start
                # from the preset and update the dict with the task parameters
                parameters = task["preset"].pop("parameters", dict())
                parameters.update(task["parameters"])
                task["parameters"] = parameters

        return tasks

    def _setup_tasks(self, global_presets, dags):
        tasks = dict()

        for dag_name, dag in dags.items():
            presets = self._get_dag_presets(dag["presets"], dag_name, global_presets)

            # Now process the tasks

            # Add tasks to the output list
            dag_tasks = self._get_dag_tasks(
                dag["tasks"], dag_name, presets, list(tasks.keys())
            )
            if dag_tasks is None:
                return
            tasks.update(dag_tasks)

        # Now that we have all tasks, check parents referenced in each task exists in the list of all tasks
        # and check if we need to load the python module
        load_python = False
        for name, task in tasks.items():
            if "parents" in task:
                missing_parents = [
                    parent
                    for parent in task.get("parents", list())
                    if parent not in tasks
                ]
                if len(missing_parents) > 0:
                    logging.error(
                        "Missing parent tasks {} referenced in task {}".format(
                            ", ".join(missing_parents), name
                        )
                    )
                    return

            if task["type"] == "python":
                load_python = True

        if load_python:
            self._setup_python_tasks_module()

        self._task_definitions = tasks



from sqlalchemy import create_engine

from .database import Database


class Sqlite(Database):
    def __init__(self, name, name_in_settings, settings):
        self.name = name
        self.name_in_settings = name_in_settings
        self.dialect = "sqlite"
        self.engine = create_engine(f'sqlite:///{settings["database"]}')
        self.create_metadata()

    #Sqlite has a different syntax for create table as which does not use ()
    def create_table_select(self, table, schema, select, replace=False, view=False):
        table = f"{schema+'.' if schema else ''}{table}"
        table_or_view = "VIEW" if view else "TABLE"

        q = ""
        if replace:
            q += f"DROP {table_or_view} IF EXISTS {table};\n"
        q += f"CREATE {table_or_view} IF NOT EXISTS {table} AS \n{select}\n;"

        return q

    #Sqlite does not support ALTER TABLE SET SCHEMA. ALTER for Sqlite can only be used to rename a table
    def move_table(self, src_table, src_schema, dst_table, dst_schema, ddl):
        drop = (
            f"DROP TABLE IF EXISTS {dst_schema+'.' if dst_schema else ''}{dst_table};"
        )
        rename = f"ALTER TABLE {src_schema+'.' if src_schema else ''}{src_table} RENAME TO {dst_table};"
        pkey_alter = ""
        if ddl is not None and 'columns' in ddl:
            if len([c for c in ddl["columns"] if c.get("primary", False)]):
                pkey_alter = f"ALTER INDEX {dst_schema+'.' if dst_schema else ''}{src_table}_pkey RENAME TO {dst_table}_pkey;"

        return "\n".join((drop, rename, pkey_alter))

    def execute(self, script):
        with self.engine.connect().execution_options(autocommit=True) as connection:
            connection.connection.executescript(script)



import logging
from pathlib import Path

from .task import Task, TaskStatus
from ..utils import yaml


class SqlTask(Task):
    def setup(self):
        self.db = self.sayn_config.default_db

        status = self._setup_file_name()
        if status != TaskStatus.READY:
            return status

        self.template = self._get_query_template()
        if self.template is None:
            return self.failed()

        try:
            self.compiled = self.template.render(**self.parameters)
        except Exception as e:
            return self.failed(f"Error compiling template\n{e}")

        return self._check_extra_fields()

    def run(self):
        logging.debug("Writting query on disk")
        self._write_query(self.compiled)
        if self.status == TaskStatus.FAILED:
            return self.failed()

        logging.debug("Running SQL")
        logging.debug(self.compiled)
        try:
            self.db.execute(self.compiled)
        except Exception as e:
            return self.failed(("Error running query", e, f"Query: {self.compiled}"))

        return self.success()

    def compile(self):
        try:
            self._write_query(self.compiled)
        except Exception as e:
            return self.failed(("Error saving query on disk", e))

        return self.success()

    # Task property methods

    def _setup_file_name(self):
        # file_name pointint to the code for the sql task
        self.file_name = self._pop_property("file_name")
        if self.file_name is None:
            return self.failed('"file_name" is a required field')
        else:
            self.file_name = self.compile_property(self.file_name)

        return TaskStatus.READY

    def _get_query_template(self):
        path = Path(self.sayn_config.sql_path, self.compile_property(self.file_name))

        if not path.is_file():
            logging.error(f"{path}: file not found")
            return

        return self.sayn_config.jinja_env.get_template(str(path))

    def _setup_destination(self):
        # Destination property indicating the table this will create
        destination = self._pop_property(
            "destination", default={"tmp_schema": None, "schema": None}
        )

        self.schema = destination.pop('schema', None)
        if self.schema is not None and isinstance(self.schema, str):
            self.schema = self.compile_property(self.schema)
        elif self.schema is not None:
            return self.failed('Optional property "schema" must be a string')

        self.tmp_schema = destination.pop('tmp_schema', None)
        if self.tmp_schema is not None and isinstance(self.tmp_schema, str):
            self.tmp_schema = self.compile_property(self.tmp_schema)
        elif self.tmp_schema is not None:
            return self.failed('Optional property "tmp_schema" must be a string')
        else:
            self.tmp_schema = self.schema

        if set(destination.keys()) == set(["table"]) and destination["table"] is not None:
            self.table = self.compile_property(destination.pop("table"))
            self.tmp_table = f"sayn_tmp_{self.table}"
        else:
            return self.failed(
                'Destination requires "table" field. Optional fields: tmp_schema and schema.'
            )

        return TaskStatus.READY

    def _setup_ddl(self, type_required=True):
        ddl = self._pop_property("ddl")
        if ddl is not None:
            if isinstance(ddl, str):
                # TODO external file not implemented
                #parsed = yaml.load(self.compile_property(ddl))
                raise ValueError('External file for ddl not implemented')
            else:
                parsed = yaml.as_document(ddl)

            self.ddl = self.db.validate_ddl(parsed, type_required=type_required)

            if self.ddl is None:
                return self.failed("Error processing DDL")
            else:
                return TaskStatus.READY

        else:
            self.ddl = None
            return TaskStatus.READY

    # Utility methods

    def _write_query(self, query, suffix=None):
        path = Path(
            self.sayn_config.compile_path,
            self.dag,
            Path(f"{self.name}{'_'+suffix if suffix is not None else ''}.sql"),
        )

        # Ensure the path exists and it's empty
        path.parent.mkdir(parents=True, exist_ok=True)
        if path.exists():
            path.unlink()

        path.write_text(str(query))



# import importlib
# from pathlib import Path

# from ..utils.singleton import singleton
from . import DatabaseError
from .postgresql import Postgresql
from .sqlite import Sqlite
from .mysql import Mysql
from .redshift import Redshift
from .snowflake import Snowflake

drivers = {
    "postgresql": Postgresql,
    "sqlite": Sqlite,
    "mysql": Mysql,
    "snowflake": Snowflake,
    "redshift": Redshift,
}


def create_all(credentials):
    return {n: create(n, c["name_in_settings"], c["settings"]) for n, c in credentials.items()}


def create(name, name_in_settings, settings):
    db_type = settings["type"]
    if db_type not in drivers:
        raise DatabaseError(f"No driver for {db_type} found")
    else:
        return drivers[db_type](name, name_in_settings, settings)



import logging

from strictyaml import (
    Bool,
    Map,
    Str,
    Optional,
    Regex,
    Any,
    MapPattern,
    UniqueSeq,
    Seq,
    Enum,
    ScalarValidator,
    as_document,
    EmptyDict,
)
from strictyaml import YAMLValidationError as ValidationError


class Identifier(Regex):
    def __init__(self):
        self._regex = r"^[a-zA-Z0-9][-_a-zA-Z0-9]+$"
        self._matching_message = "when expecting a valid identifier string"


class NotEmptyStr(Regex):
    def __init__(self):
        self._regex = r".+"
        self._matching_message = "when expecting a non-empty string"


class CaseInsensitiveStr(Str):
    def validate_scalar(self, chunk):
        return chunk.contents.lower()


class CaseInsensitiveEnum(Enum):
    def __init__(self, restricted_to):
        self._item_validator = CaseInsensitiveStr()
        for i in restricted_to:
            assert isinstance(i, str), "restricted_to must contain strings"
        assert isinstance(
            self._item_validator, ScalarValidator
        ), "item validator must be scalar too"
        self._restricted_to = [i.lower() for i in restricted_to]


def load(path):
    if not path.is_file():
        logging.exception(f"No {path.name} found")
        return
    else:
        from strictyaml import load

        try:
            return load(path.read_text())
        except ValidationError as e:
            logging.error(f"Error reading {path.name}")
            logging.error(e)
            return



import importlib
import logging

from ..utils.logger import Logger
from .task import Task
from .ignore import IgnoreTask
from .dummy import DummyTask
from .python import PythonTask
from .sql import SqlTask
from .autosql import AutoSqlTask
from .copy import CopyTask


def create_python(name, task):
    def fail_creation(name, task):
        task_object = PythonTask(name, task)
        task_object.failed()
        return task_object

    class_str = None
    if "preset" in task:
        if "preset" in task["preset"]:
            if "class" in task["preset"]["preset"]:
                class_str = task["preset"]["preset"].pop("class")

        if "class" in task["preset"]:
            class_str = task["preset"].pop("class")

    if "class" in task:
        class_str = task.pop("class")

    if class_str is None:
        logging.error('Missing required "class" field in python task')
        return fail_creation(name, task)

    module_str = ".".join(class_str.split(".")[:-1])
    class_str = class_str.split(".")[-1]

    if len(module_str) > 0:
        try:
            task_module = importlib.import_module(f"sayn_python_tasks.{module_str}")
        except Exception as e:
            logging.error(f'Error loading module "{module_str}"')
            logging.error(e)
            return fail_creation(name, task)
    else:
        task_module = importlib.import_module("sayn_python_tasks")

    try:
        klass = getattr(task_module, class_str)
    except Exception as e:
        module_file_name = (
            module_str.replace(".", "/") if len(module_str) > 0 else "__init__"
        )
        logging.error(
            f'Error importing class "{class_str}" found in "python/{module_file_name}.py"'
        )
        logging.error(e)

        return fail_creation(name, task)

    # Create object and ensure there's no extra properties
    task_object = klass(name, task)
    task_object._check_extra_fields()
    return task_object


def create_task(name, type, task, ignore):
    Logger().set_config(task=name)

    creators = {
        "ignore": IgnoreTask,
        "dummy": DummyTask,
        "sql": SqlTask,
        "autosql": AutoSqlTask,
        "copy": CopyTask,
        "python": create_python,
    }

    if ignore:
        return creators["ignore"](name, task)
    else:
        if type not in creators:
            logging.error(f'"{type}" is not a valid task type')
            return Task(name, task)
        return creators[type](name, task)



import random
import time

# Table creation queries

q_create_fighters = '''
    DROP TABLE IF EXISTS {user_prefix}logs_fighters
    ;

    CREATE TABLE {user_prefix}logs_fighters (
        event_id VARCHAR,
        fighter_id INTEGER,
        fighter_name VARCHAR
    )
    ;
'''

q_create_arenas = '''
    DROP TABLE IF EXISTS {user_prefix}logs_arenas
    ;

    CREATE TABLE {user_prefix}logs_arenas (
        event_id VARCHAR,
        arena_id INTEGER,
        arena_name VARCHAR
    )
    ;
'''

q_create_tournaments = '''
    DROP TABLE IF EXISTS {user_prefix}logs_tournaments
    ;

    CREATE TABLE {user_prefix}logs_tournaments (
        event_id VARCHAR,
        tournament_id INTEGER,
        tournament_name VARCHAR
    )
    ;
'''

q_create_battles = '''
    DROP TABLE IF EXISTS {user_prefix}logs_battles
    ;

    CREATE TABLE {user_prefix}logs_battles (
        event_id VARCHAR,
        tournament_id INTEGER,
        battle_id INTEGER,
        arena_id INTEGER,
        fighter1_id INTEGER,
        fighter2_id INTEGER,
        winner_id INTEGER
    )
    ;
'''

#Log Details

fighter_logs = [
    (1, "son goku"),
    (2, "john"),
    (3, "lucy"),
    (4, "dr. x"),
    (5, "carlos"),
    (6, "el mucho"),
    (7, "maciassito"),
    (8, "kayo"),
    (9, "vatossito"),
    (10, "reniot"),
    (11, "pepe"),
    (12, "anonymous"),
]

arena_logs = [
    (1, "earth canyon"),
    (2, "world edge"),
    (3, "namek"),
    (4, "volcanic crater"),
    (5, "underwater")
]

tournament_logs = [
    (1, "world championships"),
    (2, "tenka-ichi budokai"),
    (3, "king of the mountain")
]

def create_battle_logs(tournament_battles):
    battle_logs = []

    for tb in tournament_battles:
        tournament_id = tb["tournament_id"]
        n_battles = tb["n_battles"]

        for n in range(n_battles):
            battle_id = n + 1
            # determine fighters
            fighter1_id = random.randint(1, 12)
            fighter2_id = random.randint(1, 12)
            # loop to avoid similar fighter
            while fighter1_id == fighter2_id:
                fighter2_id = random.randint(1, 12)
            # determine arenas
            arena_id = random.randint(1, 5)
            # determine winner
            if fighter1_id == 1 or fighter2_id == 1:
                winner_id = 1
            else:
                rand = random.uniform(0, 1)
                if rand <= 0.5:
                    winner_id = fighter1_id
                else:
                    winner_id = fighter2_id

            row = (
                tournament_id,
                battle_id,
                arena_id,
                fighter1_id,
                fighter2_id,
                winner_id
            )

            battle_logs.append(row)

    return battle_logs

# Data Package

def prepare_data(tournament_battles, user_prefix=''):
    logs = {
        'fighters': {
            'create': q_create_fighters.format(user_prefix=user_prefix),
            'data': fighter_logs
        },
        'arenas': {
            'create': q_create_arenas.format(user_prefix=user_prefix),
            'data': arena_logs
        },
        'tournaments': {
            'create': q_create_tournaments.format(user_prefix=user_prefix),
            'data': tournament_logs
        },
        'battles': {
            'create': q_create_battles.format(user_prefix=user_prefix),
            'data': create_battle_logs(tournament_battles)
        },
    }

    return logs

# Log Load

def generate_load_query(log_type, log, user_prefix=''):
    event_id = time.time()

    if log_type == 'fighters':
        q_insert = '''
            INSERT INTO {user_prefix}logs_fighters VALUES('{event_id}', {fighter_id}, '{fighter_name}')
        '''.format(
            user_prefix=user_prefix,
            event_id=event_id,
            fighter_id=log[0],
            fighter_name=log[1]
        )

    elif log_type == 'arenas':
        q_insert = '''
            INSERT INTO {user_prefix}logs_arenas VALUES('{event_id}', {arena_id}, '{arena_name}')
        '''.format(
            user_prefix=user_prefix,
            event_id=event_id,
            arena_id=log[0],
            arena_name=log[1]
        )

    elif log_type == 'tournaments':
        q_insert = '''
            INSERT INTO {user_prefix}logs_tournaments VALUES('{event_id}', {tournament_id}, '{tournament_name}')
        '''.format(
            user_prefix=user_prefix,
            event_id=event_id,
            tournament_id=log[0],
            tournament_name=log[1]
        )

    elif log_type == 'battles':
        q_insert = '''
            INSERT INTO {user_prefix}logs_battles VALUES('{event_id}', {tournament_id}, {battle_id}, {arena_id}, {fighter1_id}, {fighter2_id}, {winner_id})
        '''.format(
            user_prefix=user_prefix,
            event_id=event_id,
            tournament_id=log[0],
            battle_id=log[1],
            arena_id=log[2],
            fighter1_id = log[3],
            fighter2_id = log[4],
            winner_id = log[5]
        )

    else:
        pass

    return q_insert



from enum import Enum
import logging

from ..config import Config


class TaskStatus(Enum):
    UNKNOWN = -1
    SETTING_UP = 0
    READY = 1
    EXECUTING = 2
    SUCCESS = 3
    FAILED = 4
    SKIPPED = 5
    IGNORED = 6


class Task(object):
    # Init functions
    def __init__(self, name, task):
        self.sayn_config = Config()

        self.name = name
        self.type = task.pop("type")
        self.dag = task.pop("dag")
        self.tags = task.pop("tags", list())
        self.parents = task.pop("parents", list())

        self.parameters = task.pop("parameters", dict())
        for name, value in self.parameters.items():
            self.parameters[name] = self.compile_property(self.parameters[name])

        self._task_def = task

        self.setting_up()

    def set_parents(self, tasks):
        """Replaces the parents list with the task objects"""
        self.parents = [tasks[p] for p in self.parents]

    # Properties methods

    def _check_extra_fields(self):
        # Cleanup preset
        if 'preset' in self._task_def:
            if 'preset' in self._task_def['preset']:
                if len(self._task_def['preset']['preset']) == 0:
                    # Clean nested preset if possible
                    del self._task_def['preset']['preset']

            if len(self._task_def['preset']) == 0:
                # If the nested preset is empty, clean the outer one if possible
                del self._task_def['preset']

        if len(self._task_def) > 0:
            return self.failed(
                "Invalid properties for {}: {}".format(
                    self.type, ", ".join(self._task_def.keys())
                )
            )
        else:
            return self.ready()

    def _pop_property(self, property, default=None):
        """Navigates the task definition and finds the first occurrence of the named property.
           First it checks in the task, then the preset and finally the nested preset of the preset
           (the project.yaml preset).
           Returns None if not found anywhere"""

        def merge_value(value, new_value):
            if value is None:
                return new_value
            elif not isinstance(new_value, type(value)):
                raise ValueError(type(new_value))
            elif isinstance(value, dict):
                value.update(new_value)
            elif isinstance(value, list):
                value.extend([v for v in new_value if v not in value])
            else:
                value = new_value

            return value

        value = default

        # Extract the value from the deepest level of nesting first and go outwards
        # We always remove it from the _task_def 
        if 'preset' in self._task_def:
            if 'preset' in self._task_def['preset']:
                if property in self._task_def['preset']['preset']:
                    new_value = self._task_def['preset']['preset'].pop(property)
                    value = merge_value(value, new_value)

            if property in self._task_def['preset']:
                new_value = self._task_def['preset'].pop(property)
                value = merge_value(value, new_value)

        if property in self._task_def:
            new_value = self._task_def.pop(property)
            value = merge_value(value, new_value)

        return value

    def compile_property(self, value):
        if value is None:
            return
        if isinstance(value, str):
            return Config().jinja_env.from_string(value).render(task=self)
        elif isinstance(value, list):
            return [self.compile_property(i) for i in value]
        elif isinstance(value, dict):
            return {k: self.compile_property(v) for k, v in value.items()}
        else:
            logging.error(
                "Property value type {} not supported".format(str(type(value)))
            )

    # Execution functions

    def set_current(self):
        self.sayn_config.set_current_task(self.name)

    def should_run(self):
        # Initially all tasks should run, except IgnoreTask
        return True

    def can_run(self):
        if self.status != TaskStatus.READY:
            return False
        for p in self.parents:
            if p.status not in (TaskStatus.IGNORED, TaskStatus.SUCCESS):
                return False
        return True

    # API

    def setup(self):
        return self.failed()

    def run(self):
        return self.failed()

    def compile(self):
        return self.failed()

    # Status functions

    def setting_up(self):
        self.status = TaskStatus.SETTING_UP
        return self.status

    def ready(self):
        self.status = TaskStatus.READY
        return self.status

    def executing(self):
        self.status = TaskStatus.EXECUTING
        return self.status

    def success(self):
        self.status = TaskStatus.SUCCESS
        return self.status

    def failed(self, messages=None):
        if messages is not None:
            if isinstance(messages, str):
                messages = [messages]
            for message in messages:
                logging.error(message)

        self.status = TaskStatus.FAILED
        return self.status

    def skipped(self):
        self.status = TaskStatus.SKIPPED
        return self.status



import logging
import coloredlogs
from pathlib import Path

from .singleton import singleton


@singleton
class Logger:
    def __init__(self, run_id=None, debug=False, log_file=Path("logs/sayn.log")):
        self.run_id = run_id
        self.level = logging.DEBUG if debug else logging.INFO
        logging.getLogger().setLevel(self.level)
        self.log_file = log_file
        self.sayn_project = Path().cwd().parts[-1]

        self.progress = None
        self.task_name = None
        self.stage_name = None

        self.set_console_logger()
        self.set_file_logger(log_file)

    def set_config(self, **kwargs):
        if 'stage' in kwargs:
            self.stage_name = kwargs['stage']

        if 'task' in kwargs:
            self.task_name = kwargs['task']

        if 'progress' in kwargs:
            self.progress = kwargs['progress']

        self._set_console_formatter()
        self._set_file_formatter()

    def set_debug(self):
        self.level = logging.DEBUG
        logging.getLogger().setLevel(self.level)
        self.console_handler.setLevel(self.level)
        if self.file_handler is not None:
            self.file_handler.setLevel(self.level)

    def set_console_logger(self):
        self.console_handler = logging.StreamHandler()
        self.console_handler.setLevel(self.level)
        self._set_console_formatter()
        logging.getLogger().addHandler(self.console_handler)

    def set_file_logger(self, log_file=None):
        if log_file is not None:
            self.log_file = log_file

        if self.log_file is None:
            self.file_handler = None
        else:
            # Create folder if it doesn't exists
            if not self.log_file.parent.exists():
                self.log_file.parent.mkdir(parents=True)

            self.file_handler = logging.FileHandler(self.log_file)
            self.file_handler.setLevel(self.level)

            self._set_file_formatter()
            logging.getLogger().addHandler(self.file_handler)

    def _set_console_formatter(self):
        fmt_string = "[%(asctime)s] "

        if self.stage_name is not None:
            fmt_string += f"{self.stage_name}|"

        if self.progress is not None:
            fmt_string += f"{self.progress}%%|"

        fmt_string += "%(levelname)s|"

        if self.task_name is not None:
            fmt_string += f"{self.task_name}|"

        fmt_string +="%(message)s"

        formatter = coloredlogs.ColoredFormatter(
            fmt=fmt_string,
            field_styles={
            #    "asctime": {}
            #    "stage_name": {"bold": True, "color": "blue"},
            #    "progress": {"bold": True, "color": "blue"},
            },
            level_styles={
                "debug": {"color": "blue"},
                "info": {},
                "warning": {"color": "yellow"},
                "error": {"color": "red"},
                "critical": {"color": "red"},
                #"exception": {"color": "red"},
            },
        )
        self.console_handler.setFormatter(formatter)

    def _set_file_formatter(self):
        if self.file_handler is not None:
            fmt_string = f"{self.run_id}|{self.sayn_project}|" + "%(asctime)s|"

            if self.stage_name is not None:
                fmt_string += f"{self.stage_name}|"

            if self.progress is not None:
                fmt_string += f"{self.progress}%%|"

            fmt_string += "%(levelname)s|"

            if self.task_name is not None:
                fmt_string += f"{self.task_name}|"

            fmt_string +="%(message)s"

            formatter = logging.Formatter(fmt_string, "%Y-%m-%d %H:%M:%S",)
            self.file_handler.setFormatter(formatter)

    def _set_formatters(self):
        self._set_console_formatter()
        self._set_file_formatter()



import sqlite3
import logging
from .utils.log_creator import prepare_data, generate_load_query
from sayn import PythonTask

class LoadData(PythonTask):
    def setup(self):
       err = False

       # we use this list to control how many battles we want per tournament
       tournament_battles = [
           {"tournament_id": 1, "n_battles": 1000},
           {"tournament_id": 2, "n_battles": 250},
           {"tournament_id": 3, "n_battles": 500},
       ]

       try:
           self.data_to_load = prepare_data(tournament_battles)
       except Exception as e:
           err = True
           logging.error(e)

       if err:
           return self.failed()
       else:
           return self.ready()

    def run(self):

        # load the logs
        for log_type, log_details in self.data_to_load.items():
            # create table
            logging.info('Creating table: {log_type}.'.format(log_type=log_type))

            self.default_db.execute(log_details['create'])

            # load logs
            logging.info('Loading logs: {log_type}.'.format(log_type=log_type))
            logs = log_details['data']

            for log in logs:
                q_insert = generate_load_query(log_type, log)

                self.default_db.execute(q_insert)

            # done
            logging.info('Done: {log_type}.'.format(log_type=log_type))

        return self.success()



from .database import Database


class Mysql(Database):
    def __init__(self, name, name_in_settings, settings):
        self.dialect = "postgresql"
        connection_details = settings
        connection_details.pop("type")
        super().__init__(name, name_in_settings, connection_details)

    def execute(self, script):
        with self.engine.connect().execution_options(autocommit=True) as connection:
            connection.execute(script, multi=True)



import os
from pathlib import Path
import shutil


def sayn_init(sayn_project_name):
    src = Path(os.path.dirname(os.path.abspath(__file__)), 'sayn_project_base').absolute()
    dst = Path(sayn_project_name).absolute()

    print('Creating SAYN base project at the following location: {dst}...'.format(dst=dst))
    try:
        shutil.copytree(src, dst)

        # rename the sample_settings.yaml to settings.yaml
        os.rename(Path(dst, 'sample_settings.yaml').absolute(), Path(dst, 'settings.yaml').absolute())

        print('SAYN base project created.')
        print('For more information about SAYN visit our documentation:')
        print('https://173tech.github.io/sayn/')
    except OSError as e:
        print(e)



from .task import Task, TaskStatus
from .creator import create_task

from .ignore import IgnoreTask
from .dummy import DummyTask
from .python import PythonTask
from .sql import SqlTask
from .autosql import AutoSqlTask
from .copy import CopyTask



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sayn',
 'sayn.commands',
 'sayn.database',
 'sayn.start_project',
 'sayn.start_project.sayn_project_base.python',
 'sayn.start_project.sayn_project_base.python.utils',
 'sayn.tasks',
 'sayn.utils']

package_data = \
{'': ['*'],
 'sayn.start_project': ['sayn_project_base/*',
                        'sayn_project_base/dags/*',
                        'sayn_project_base/sql/*']}

install_requires = \
['Click>=7.0,<8.0',
 'coloredlogs>=14.0,<15.0',
 'jinja2>=2.10.3,<3.0.0',
 'sqlalchemy>=1.3.12,<2.0.0',
 'strictyaml>=1.0.6,<2.0.0']

extras_require = \
{':extra == "graphviz"': ['graphviz>=0.14,<0.15'],
 'postgresql': ['psycopg2>=2.8.5,<3.0.0'],
 'snowflake': ['snowflake-connector-python>=2.0.3,<3.0.0',
               'snowflake-sqlalchemy>=1.1.18,<2.0.0']}

entry_points = \
{'console_scripts': ['sayn = sayn.commands.sayn_powers:cli']}

setup_kwargs = {
    'name': 'sayn',
    'version': '0.2.1',
    'description': 'Data-modelling and processing framework for automating Python and SQL tasks',
    'long_description': '<img\n  src="https://173-static-files.s3.eu-west-2.amazonaws.com/sayn_docs/logos/sayn_logo.png"\n  alt="SAYN logo"\n  style="width: 50%; height: 50%;"\n/>\n\n#\n\nSAYN is a data-modelling and processing framework for automating Python and SQL tasks. It enables analytics teams to build robust data infrastructures in minutes.\n\n *Status: SAYN is under active development so some changes can be expected.*\n\n## Use Cases\n\nSAYN can be used for multiple purposes across the analytics workflow:\n\n* Data extraction: complement tools such as Fivetran or Stitch with customised extraction processes.\n* Date modelling: transform raw data in your data warehouse.\n* Date science: integrate and execute data science models.\n\n## Key Features\n\nSAYN has the following key features:\n\n* YAML based creation of [DAGs](https://173tech.github.io/sayn/dags/) (Direct Acyclic Graph). This means all analysts, including non Python proficient ones, can contribute to building ETL processes.\n* [SQL SELECT statements](https://173tech.github.io/sayn/tasks/autosql/): turn your queries into managed tables and views automatically.\n* [Jinja parameters](https://173tech.github.io/sayn/parameters/): switch easily between development and product environment and other tricks with Jinja templating.\n* [Python tasks](https://173tech.github.io/sayn/tasks/python/): use Python scripts to complement your extraction and loading layer and build data science models.\n* Multiple [databases](https://173tech.github.io/sayn/databases/overview/) supported.\n* and much more... See the [Documentation](https://173tech.github.io/sayn/).\n\n## Design Principles\n\nSAYN is designed around three core principles:\n\n* **Simplicity**: data models and processes should be easy to create, scale and maintain. So your team can focus on data transformation instead of writing processes. SAYN orchestrates all your tasks systematically and provides a lot of automation features.\n* **Flexibility**: the power of data is unlimited and so should your tooling. SAYN currently supports both SQL and Python so your analysts can choose the most optimal solution for each process.\n* **Centralisation**: all analytics code should live in one place, making your life easier and allowing dependencies throughout the whole analytics process.\n\n## Quick Start\n\n```bash\n$ pip install sayn\n$ sayn init test_sayn\n$ cd test_sayn\n$ sayn run\n```\n\nThis is it! You completed your first SAYN run on the example project. Continue with the [Tutorial: Part 1](https://173tech.github.io/sayn/tutorials/tutorial_part1/) which will give you a good overview of SAYN\'s true power!\n\n## Support\n\nIf you need any help with SAYN, or simply want to know more, please contact the team at <sayn@173tech.com>.\n\n## License\n\nSAYN is open source under the [Apache 2.0](https://www.apache.org/licenses/LICENSE-2.0) license.\n\n---\n\nMade with :heart: by [173tech](https://www.173tech.com).\n',
    'author': 'Robin Watteaux',
    'author_email': 'robin@173tech.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://173tech.github.io/sayn',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from itertools import groupby
import logging

from sqlalchemy import create_engine, MetaData, Table
from sqlalchemy.sql import select, func, or_, text

from . import DatabaseError
from ..utils import yaml


class Database:
    def __init__(self, name, name_in_settings, connection_details):
        self.name = name
        self.name_in_settings = name_in_settings
        self.engine = create_engine(f"{self.dialect}://", **connection_details)
        self.create_metadata()

    def validate_ddl(self, ddl, type_required=True):
        schema = yaml.Map(
            {
                yaml.Optional("columns"): yaml.Seq(
                    yaml.Map(
                        {
                            "name": yaml.NotEmptyStr(),
                            "type"
                            if type_required
                            else yaml.Optional("type"): yaml.NotEmptyStr(),
                            yaml.Optional("primary"): yaml.Bool(),
                            yaml.Optional("not_null"): yaml.Bool(),
                            yaml.Optional("unique"): yaml.Bool(),
                            yaml.Optional("indexes"): yaml.UniqueSeq(
                                yaml.NotEmptyStr()
                            ),
                        }
                    )
                ),
                yaml.Optional("permissions"): yaml.MapPattern(
                    yaml.NotEmptyStr(), yaml.NotEmptyStr()
                ),
            }
        )

        try:
            ddl.revalidate(schema)
        except Exception as e:
            logging.error(e)
            return

        return ddl.data

    def connect(self):
        return self.engine.connect()

    def execute(self, script):
        with self.engine.connect().execution_options(autocommit=True) as connection:
            connection.execute(script)

    def select(self, query, params=None):
        if params is not None:
            res = self.engine.execute(query, **params)
        else:
            res = self.engine.execute(query)

        return [dict(zip(res.keys(), r)) for r in res.fetchall()]

    def load_data(self, table, schema, data):
        table_def = self.get_table(table, schema)
        with self.engine.connect().execution_options(autocommit=True) as connection:
            connection.execute(table_def.insert().values(data))

    def create_metadata(self):
        self.metadata = MetaData(self.engine)

    def refresh_metadata(self, only=None, schema=None):
        self.metadata.reflect(only=only, schema=schema, extend_existing=True)

    # ETL steps

    def get_table(self, table, schema, columns=None, required_existing=False):
        """Create a SQLAlchemy Table object. If columns is not None, fills up columns or checks the columns are present"""
        table_def = Table(table, self.metadata, schema=schema, extend_existing=True)

        if table_def.exists():
            self.refresh_metadata(only=[table], schema=schema)
            if columns is not None:
                cols_in_table = set([c.name for c in table_def.columns])
                cols_requested = set(
                    [c.name if not isinstance(c, str) else c for c in columns]
                )

                if len(cols_requested - cols_in_table) > 0:
                    logging.error(
                        f"Missing columns \"{', '.join(cols_requested - cols_in_table)}\" in table \"{table_def.name}\""
                    )
                    return

                if len(cols_in_table - cols_requested) > 0:
                    for column in cols_in_table - cols_requested:
                        table_def._columns.remove(table_def.columns[column])
        elif required_existing:
            return
        elif columns is not None:
            for column in columns:
                table_def.append_column(column.copy())

        return table_def

    def table_exists(self, table, schema, with_columns=None):
        table_def = self.get_table(table, schema, columns=with_columns)
        if table_def is not None:
            return table_def.exists()
        else:
            return False

    def create_table_select(self, table, schema, select, replace=False, view=False):
        table = f"{schema+'.' if schema else ''}{table}"
        table_or_view = "VIEW" if view else "TABLE"

        q = ""
        if replace:
            q += f"DROP {table_or_view} IF EXISTS {table};\n"
        q += f"CREATE {table_or_view} IF NOT EXISTS {table} AS (\n{select}\n);"

        return q

    def create_table_ddl(self, table, schema, ddl, replace=False):
        table_name = table
        table = f"{schema+'.' if schema else ''}{table_name}"
        columns = "\n    , ".join(
            [
                (
                    f'{c["name"]} {c["type"]}'
                    f'{" PRIMARY KEY" if c.get("primary", False) else ""}'
                    f'{" NOT NULL" if c.get("not_null", False) else ""}'
                )
                for c in ddl["columns"]
            ]
        )
        indexes = {
            idx: [c[1] for c in cols]
            for idx, cols in groupby(
                sorted(
                    [
                        (idx, c["name"])
                        for c in ddl["columns"]
                        if "indexes" in c
                        for idx in c["indexes"]
                    ]
                ),
                lambda x: x[0],
            )
        }

        q = ""
        if replace:
            q += f"DROP TABLE IF EXISTS {table};\n"
        q += f"CREATE TABLE IF NOT EXISTS {table} (\n      {columns}\n);\n"
        q += "\n".join(
            [
                f"CREATE INDEX IF NOT EXISTS {table_name}_{name} ON {table}({', '.join(cols)});"
                for name, cols in indexes.items()
            ]
        )

        return q

    def grant_permissions(self, table, schema, ddl):
        return "\n".join(
            [
                f"GRANT {priv} ON {schema+'.' if schema else ''}{table} TO \"{role}\";"
                for role, priv in ddl.items()
            ]
        )

    def drop_table(self, table, schema, view=False):
        table = f"{schema+'.' if schema else ''}{table}"
        table_or_view = "VIEW" if view else "TABLE"
        return f"DROP {table_or_view} IF EXISTS {table};"

    def insert(self, table, schema, select):
        table = f"{schema+'.' if schema else ''}{table}"
        return f"INSERT INTO {table} (\n{select}\n);"

    def move_table(self, src_table, src_schema, dst_table, dst_schema, ddl):
        drop = (
            f"DROP TABLE IF EXISTS {dst_schema+'.' if dst_schema else ''}{dst_table};"
        )
        rename = f"ALTER TABLE {src_schema+'.' if src_schema else ''}{src_table} RENAME TO {dst_table};"
        if dst_schema is not None and dst_schema != src_schema:
            change_schema = f"ALTER TABLE {src_schema+'.' if src_schema else ''}{dst_table} SET SCHEMA {dst_schema};"
        else:
            change_schema = ""

        idx_alter = []
        if ddl is not None and "columns" in ddl:
            # Change primary key name
            if len([c for c in ddl["columns"] if c.get("primary", False)]):
                idx_alter.append(
                    f"ALTER INDEX {dst_schema+'.' if dst_schema else ''}{src_table}_pkey RENAME TO {dst_table}_pkey;"
                )

            # Change index names
            for idx in [
                idx for c in ddl["columns"] for idx in c.get("indexes", list())
            ]:
                idx_alter.append(
                    f"ALTER INDEX {dst_schema+'.' if dst_schema else ''}{src_table}_{idx} RENAME TO {dst_table}_{idx};"
                )

        return "\n".join([drop, rename, change_schema] + idx_alter)

    def merge_tables(self, src_table, src_schema, dst_table, dst_schema, delete_key):
        dst = f"{dst_schema+'.' if dst_schema else ''}{dst_table}"
        src = f"{src_schema+'.' if src_schema else ''}{src_table}"

        delete = (
            f"DELETE FROM {dst}\n"
            f" WHERE EXISTS (SELECT *\n"
            f"                 FROM {src}\n"
            f"                WHERE {src}.{delete_key} = {dst}.{delete_key});"
        )
        insert = f"INSERT INTO {dst} SELECT * FROM {src};"
        drop = f"DROP TABLE {src};"
        return "\n".join((delete, insert, drop))

    def copy_from_table(
        self, src, dst_table, dst_schema=None, columns=None, incremental_column=None
    ):
        # 1. Check source table
        if not src.exists():
            raise DatabaseError(f"Source table {src.fullname} does not exists")

        if columns is None:
            columns = [c.name for c in src.columns]
            if incremental_column is not None and incremental_column not in src.columns:
                raise DatabaseError(
                    f"Incremental column {incremental_column} not in source table {src.fullname}"
                )
        else:
            if incremental_column is not None and incremental_column not in columns:
                columns.append(incremental_column)

            for column in columns:
                if column not in src.columns:
                    raise DatabaseError(
                        f"Specified column {column} not in source table {src.fullname}"
                    )

        # 2. Get incremental value
        table = self.get_table(dst_table, dst_schema, columns=columns)

        if incremental_column is not None and not dst_table.exists():
            if incremental_column not in table.columns:
                raise DatabaseError(
                    f"Incremental column {incremental_column} not in destination table {table.fullname}"
                )
            incremental_value = (
                select([func.max(table.c[incremental_column])])
                .where(table.c[incremental_column] != None)
                .execute()
                .fetchone()[0]
            )
        elif table is None:
            # Create table
            table = Table(dst_table, self.metadata, schema=dst_schema)
            for column in columns:
                table.append_column(src.columns[column].copy())
            table.create()
            incremental_value = None
        else:
            incremental_value = None

        # 3. Get data
        if incremental_value is None:
            where_cond = True
        else:
            where_cond = or_(
                src.c[incremental_column] == None,
                src.c[incremental_column] > incremental_value,
            )
        query = select([src.c[c] for c in columns]).where(where_cond)
        data = query.execute().fetchall()

        # 4. Load data
        return self.load_data(
            [dict(zip([str(c.name) for c in query.c], row)) for row in data], table
        )

    def get_max_value(self, table, schema, column):
        table_def = self.get_table(table, schema)
        return select([func.max(table_def.c[column])]).where(
            table_def.c[column] != None
        )

    def get_data(self, table, schema, columns, incremental_key=None):
        src = self.get_table(table, schema, columns)
        q = select([src.c[c] for c in columns])
        if incremental_key is not None:
            q = q.where(
                or_(
                    src.c[incremental_key] == None,
                    src.c[incremental_key] > text(":incremental_value"),
                )
            )
        return q



import logging

from .task import TaskStatus
from .sql import SqlTask


class CopyTask(SqlTask):
    def setup(self):
        self.db = self.sayn_config.default_db

        status = self._setup_ddl(type_required=False)
        if self.ddl is None:
            # TODO full copy of table when ddl not specified
            return self.failed("DDL is required for copy tasks")

        status = self._setup_source()
        if status != TaskStatus.READY:
            return status

        status = self._setup_destination()
        if status != TaskStatus.READY:
            return status

        status = self._setup_incremental()
        if status != TaskStatus.READY:
            return status

        status = self._setup_table_columns()
        if status != TaskStatus.READY:
            return status

        status = self._setup_sql()
        if status != TaskStatus.READY:
            return status

        return self.ready()

    def run(self):
        logging.debug("Writting query on disk")
        status = self._write_queries()
        if status != TaskStatus.SUCCESS:
            return self.failed()

        logging.debug("Running SQL")
        try:
            step = "last_incremental_value"
            last_incremental_value = None
            if self.compiled[step] is not None:
                res = self.db.select(self.compiled[step])
                res = list(res[0].values())[0]
                if res is not None:
                    last_incremental_value = {'incremental_value': res}

            step = "get_data"
            data = self.source_db.select(self.compiled[step], last_incremental_value)

            if len(data) == 0:
                logging.debug('Nothing to load')
                return self.success()

            step = "create_load_table"
            if self.compiled[step] is not None:
                self.db.execute(self.compiled[step])

            step = "load_data"
            self.db.load_data(self.load_table, self.load_schema, data)

            step = "finish"
            if self.compiled[step] is not None:
                self.db.execute(self.compiled[step])
        except Exception as e:
            logging.error(f'Error running step "{step}"')
            logging.error(e)
            if step != "load_data":
                logging.debug(f"Query: {self.compiled[step]}")
            return self.failed()

        return self.success()

    def compile(self):
        status = self._write_queries()
        if status != TaskStatus.SUCCESS:
            return self.failed()
        else:
            return self.success()

    def _write_queries(self):
        try:

            step = "last_incremental_value"
            if self.compiled[step] is not None:
                self._write_query(self.compiled[step], suffix=step)

            step = "get_data"
            self._write_query(self.compiled[step], suffix=step)

            step = "create_load_table"
            if self.compiled[step] is not None:
                self._write_query(self.compiled[step], suffix=step)

            step = "finish"
            if self.compiled[step] is not None:
                self._write_query(self.compiled[step], suffix=step)
        except Exception as e:
            return self.failed((f'Error saving query "{step}" on disk', e))

        return TaskStatus.SUCCESS

    # Task property methods

    def _setup_incremental(self):
        # Type of materialisation
        self.delete_key = self._pop_property("delete_key")
        self.incremental_key = self._pop_property("incremental_key")
        
        if (self.delete_key is None) != (self.incremental_key is None):
            return self.failed(
                f'Incremental copy requires both "delete_key" and "incremental_key"'
            )

        return TaskStatus.READY

    def _setup_source(self):
        # Source property indicating the table this will create
        source = self._pop_property(
            "source", default={"schema": None}
        )

        self.schema = source.pop('schema', None)
        if self.schema is not None and isinstance(self.schema, str):
            self.schema = self.compile_property(self.schema)
        else:
            return self.failed('Optional property "schema" must be a string')

        if set(source.keys()) != set(["db", "table"]) or source['db'] is None or source['table'] is None:
            return self.failed(
                'Source requires "table" and "db" fields. Optional field: "schema".'
            )
        else:
            source_db_name = self.compile_property(source.pop("db"))
            self.source_table = self.compile_property(source.pop("table"))

        if source_db_name not in self.sayn_config.dbs:
            return self.failed(f'{source_db_name} is not a valid vallue for "db" in "source"')
        else:
            self.source_db = self.sayn_config.dbs[source_db_name]

        return TaskStatus.READY

    def _setup_table_columns(self):
        self.source_table_def = self.source_db.get_table(
            self.source_table,
            self.source_schema,
            columns=[c["name"] for c in self.ddl["columns"]],
            required_existing=True
        )

        if self.source_table_def is None or not self.source_table_def.exists():
            return self.failed(
                (
                    f"Table \"{self.source_schema+'.' if self.source_schema is not None else ''}{self.source_table}\""
                    f" does not exists or columns don't match with DDL specification"
                )
            )

        # Fill up column types from the source table
        for column in self.ddl["columns"]:
            if "type" not in column:
                column["type"] = self.source_table_def.columns[
                    column["name"]
                ].type.compile()

        return TaskStatus.READY

    # Utility methods

    def _setup_sql(self):
        self.destination_table_def = self.db.get_table(
            self.table, self.schema, columns=self.source_table_def.columns
        )

        if self.destination_table_def is None:
            return self.failed('Error detecting destination table')

        if not self.destination_table_def.exists():
            # Full load:
            # Create final table directly and load there
            self.load_table = self.table
            self.load_schema = self.schema
            create_load_table = self.db.create_table_ddl(
                self.load_table, self.load_schema, self.ddl
            )
            self.compiled = {
                "last_incremental_value": None,
                "get_data": self.source_db.get_data(
                    self.source_table,
                    self.source_schema,
                    [c["name"] for c in self.ddl["columns"]],
                ),
                "create_load_table": f"-- Create table\n{create_load_table}",
                "finish": None,
            }

        elif self.incremental_key is None:
            # Table exists and it's not an incremental load:
            # Create temporary table, load there and then move
            self.load_table = self.tmp_table
            self.load_schema = self.tmp_schema
            create_load_table = self.db.create_table_ddl(
                self.load_table, self.load_schema, self.ddl, replace=True
            )
            self.compiled = {
                "last_incremental_value": None,
                "get_data": self.source_db.get_data(
                    self.source_table,
                    self.source_schema,
                    [c["name"] for c in self.ddl["columns"]],
                ),
                "create_load_table": f"-- Create temporary table\n{create_load_table}",
                "finish": self.db.move_table(
                    self.load_table,
                    self.load_schema,
                    self.table,
                    self.schema,
                    self.ddl,
                ),
            }
        elif self.delete_key is not None:
            # Incremental load:
            # Create temporary table, load there and then merge
            self.load_table = self.tmp_table
            self.load_schema = self.tmp_schema
            create_load_table = self.db.create_table_ddl(
                self.load_table, self.load_schema, self.ddl, replace=True
            )
            self.compiled = {
                "last_incremental_value": self.db.get_max_value(
                    self.table, self.schema, self.incremental_key
                ),
                "get_data": self.source_db.get_data(
                    self.source_table,
                    self.source_schema,
                    [c["name"] for c in self.ddl["columns"]],
                    incremental_key=self.incremental_key,
                ),
                "create_load_table": f"-- Create temporary table\n{create_load_table}",
                "finish": self.db.merge_tables(
                    self.load_table,
                    self.load_schema,
                    self.table,
                    self.schema,
                    self.delete_key,
                ),
            }
        else:
            raise ValueError("Sayn error in copy")

        return TaskStatus.READY



class SingletonMeta(type):
    _instance = None

    def __call__(self):
        if self._instance is None:
            self._instance = super().__call__()
        return self._instance


def singleton(class_):
    instances = {}
    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]
    return getinstance




import logging

from sqlalchemy import create_engine

from .database import Database


class Snowflake(Database):
    def __init__(self, name, name_in_settings, settings):
        self.dialect = "snowflake"
        connection_details = settings
        connection_details.pop("type")

        self.name = name
        self.name_in_settings = name_in_settings

        from snowflake.sqlalchemy import URL
        
        for logger_name in ['snowflake.connector', 'botocore', 'boto3']:
            logger = logging.getLogger(logger_name)
            logger.setLevel(logging.WARNING)

        self.engine = create_engine(URL(**connection_details["connect_args"]))

        self.create_metadata()

    def execute(self, script):
        conn = self.engine.connect()
        conn.connection.execute_string(script)
        conn.connection.close()
        # with self.engine.connect().execution_options(autocommit=1) as connection:
        #     connection.connection.execute_string(script)

    # def _get_schema(self, schema):
    #     if schema is not None and "." in schema:
    #         logging.warning('Changed DB to '+schema.split('.')[0])
    #         self.engine.execute(f"USE {schema.split('.')[0]}")
    #         return schema.split('.')[1]

    #     return schema

    # def refresh_metadata(self, only=None, schema=None):
    #     if schema is not None and "." in schema:
    #         logging.warning('Changed DB to '+schema.split('.')[0])
    #         self.engine.execute(f"USE {schema.split('.')[0]}")
    #     self.metadata.reflect(only=only, schema=schema, extend_existing=True)

    # def get_table(self, table, schema, columns=None, required_existing=False):
    #     """Create a SQLAlchemy Table object. If columns is not None, fills up columns or checks the columns are present"""
    #     schema = self._get_schema(schema)
    #     return super(Snowflake, self).get_table(table, schema, columns, required_existing=required_existing)

    def move_table(self, src_table, src_schema, dst_table, dst_schema, ddl):
        drop = (
            f"DROP TABLE IF EXISTS {dst_schema+'.' if dst_schema else ''}{dst_table};"
        )
        rename = f"ALTER TABLE {src_schema+'.' if src_schema else ''}{src_table} RENAME TO {dst_schema+'.' if dst_schema else ''}{dst_table};"

        pkey_alter = ""
        if ddl is not None and 'columns' in ddl:
            if len([c for c in ddl["columns"] if c.get("primary", False)]):
                pkey_alter = f"ALTER INDEX {dst_schema+'.' if dst_schema else ''}{src_table}_pkey RENAME TO {dst_table}_pkey;"

        return "\n".join((drop, rename, pkey_alter))




import logging

from .task import Task


class PythonTask(Task):
    def __init__(self, name, task):
        super(PythonTask, self).__init__(name, task)

        # Add some convenient properties
        self.default_db = self.sayn_config.default_db

    def setup(self):
        logging.debug("Setting up Python Task")
        return self.ready()

    def compile(self):
        logging.debug("PythonTask compiling")
        return self.success()

    def run(self):
        logging.debug("PythonTask running.")
        return self.success()



class DatabaseError(Exception): pass

from .creator import create_all
from .database import Database
from .postgresql import Postgresql
from .redshift import Redshift
from .snowflake import Snowflake
from .mysql import Mysql
from .sqlite import Sqlite
